/*
 * macro.c, part of e
 * read COPYING now for more important info.
 *
 * macro-handling
 *
 * note: ALL functions return 0 on error!
 *	 a returnvalue of 1 means success unless stated otherwise
 */


#include <stdio.h>
#include <curses.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "e.h"


int parse_statusbar_definition( char *, char * );


struct str_macro {
	char	*name;
	int 	(* function)( char * );
};

typedef struct str_macro 	STR_MACRO;

STR_MACRO	macros[] = {
		{ "exit", 		do_exit 		},
		
		{ "loadfile",		do_loadfile 		},
		{ "closefile",		do_closefile		},
		{ "savefile",		do_savefile		},
		
		{ "nextfile",		do_nextfile		},
		{ "prevfile",		do_prevfile		},
		
		{ "newwin",		do_newwin		},
		{ "closewin",		do_closewin		},
		
		{ "refresh",		do_refresh		},
		{ "refreshline",	do_refreshline		},
		{ "refreshcursor",	do_refreshcursor	},
		{ "refreshstatusbar",	do_refreshstatusbar	},
		
		{ "showstatus",		do_showstatus		},
		
		{ "delline",		do_delline		},
		{ "delleft",		do_delleft		},
		{ "delright",		do_delright		},
		
		{ "up",			do_up			},
		{ "down",		do_down			},
		{ "left",		do_left			},
		{ "right",		do_right		},
		{ "startofline",	do_startofline		},
		{ "endofline",		do_endofline		},
		{ "pageup",		do_pageup		},
		{ "pagedown",		do_pagedown		},
		
		{ "cursoroffset",	do_cursoroffset		},
		
		{ "insert",		do_insert		},

		{ NULL,			NULL 			}
		};


/*
 * tries to interpret and run a macro
 */
int macro_run( const char *format, ... )
{
	int		i;
	char		*macro;
	char		mac[MAX_STRINGSIZE];
	va_list		ap;

	va_start( ap, format );
	
	vsnprintf( mac, MAX_STRINGSIZE, format, ap );

	macro = mac;
	
	while( macro[0] ) {
		for( i = 0; macros[i].name; i++ )
			if( !strncasecmp( macro, macros[i].name, strlen( macros[i].name ) ) && ( !macro[strlen( macros[i].name )] || isspace( macro[strlen( macros[i].name )] ) ) ) {
				e_log( "macro started: %s", macro );
				if( !macros[i].function( macro + strlen( macros[i].name ) ) ) {
					e_log( "macro failed: %s\n", macro );
					return 0;
				}
				macro = seeknextline( macro );
				break;
			}
		if( !macros[i].name ) {
			e_log( "unknown macro: %s\n", macro );
			e_msg( "unknown macro: %s", macro );
			refresh();
			return 0;
		}
	}

	e_log( "macro completed: %s\n", mac );
	
	return 1;
}


/*
 * exits e
 *
 * returns: never returns
 */
int do_exit( char *arg )
{
	int		errorlevel;

	errorlevel = atoi( stripspaces( arg ) );
	e_log( "shutting down with errorlevel %d...\n", errorlevel );
	exit( errorlevel );
}


/*
 * tries to load a file
 */
int do_loadfile( char *arg )
{
	FILE		*in;
	STR_FILE	*newfile;
	char		filename[MAX_PATHSIZE];
	char		tmpline[MAX_LINESIZE];
	STR_LINE	*newline;
	STR_LINE	*prevline = NULL;
	STR_FILE	*oldfile;

	arg = stripspaces( arg );
	toeol( filename, arg, 512 );

	if( !( in = fopen( filename, "rb" ) ) ) return 0;
	
	fseek( in, 0, SEEK_END );
	
	if( !( newfile = (STR_FILE *) e_alloc( sizeof( STR_FILE ) ) ) ) {
		fclose( in );
		return 0;
	}

	newfile->size = ftell( in );

	fseek( in, 0, SEEK_SET );

	oldfile = e->curwin->file;
	e->curwin->file = newfile;
	tmpline[0] = '\n';
	tmpline[1] = 0;

	while( !feof( in ) ) {
		if( !( newline = (STR_LINE *) e_alloc( sizeof( STR_LINE ) ) ) ) break;
		newline->offset = ftell( in );
		if( !fgets( tmpline, MAX_LINESIZE, in ) ) {
			if( !feof( in ) ) {
				e_msg( "error reading file '%s'!", arg );
				break;
			} else tmpline[0] = 0;  
		}
		if( strlen( tmpline ) == MAX_LINESIZE - 1 && !feof( in ) ) {
			e_msg( "line too long!" );
			break;
		}
		newline->size = strlen( tmpline );
		
		newline->maxsize = ( (int) ( ( newline->size + 1 ) / ALLOC_SIZE ) + 1 ) * ALLOC_SIZE;
		
		if( ( newline->buf = (char *) e_alloc( newline->maxsize ) ) ) {
			strcpy( newline->buf, tmpline );
			newline->bufpos = 0;
		} else newline->bufpos = 1;
		
		if( prevline ) {
			prevline->next = newline;
			newline->prev = prevline;
		} else {
			newfile->firstline = newline;
			newline->prev = NULL;
		}
		newline->next = NULL;
		prevline = newline;
	}
	
	if( !feof( in ) ) {
		fclose( in );
		if( !do_closefile( "" ) ) e_msg( "error closing file!" );
		e->curwin->file = oldfile;
		return 0;
	}
	
	if( oldfile ) {
		if( oldfile->next ) {
			newfile->next = oldfile->next;
			newfile->next->prev = newfile;
		} else newfile->next = NULL;
		oldfile->next = newfile;
		newfile->prev = oldfile;
	} else {
		newfile->prev = NULL;
		newfile->next = NULL;
	}

	newfile->orig = in;
	newfile->tabsize = e->tabsize;
	newfile->curline = newfile->firstline;
	newfile->line = 0;
	newfile->col = 0;
	newfile->topline = 0;
	newfile->leftcol = 0;
	strncpy( newfile->name, filename, MAX_PATHSIZE - 1 );
	newfile->name[MAX_PATHSIZE - 1] = '\0';
	e->curwin->curline = newfile->curline;

	do_refresh( "" );

	return 1;
}


/*
 * tries to close the current file
 */
int do_closefile( char *arg )
{
	STR_WIN		*oldwin, *w, *w2;
	STR_FILE	*kill;
	STR_LINE	*l, *l2;
	
	if( !( kill = e->curwin->file ) ) return 1;
	
	for( l = kill->firstline; l; ) {
		if( l->next ) l2 = l->next;
		e->curwin->curline = l;
		if( !do_delline( "" ) ) return 0;
		l = l2;
	}

	free( kill->name );
	
	if( kill->prev ) {
		if( kill->next ) {
			kill->prev->next = kill->next;
			kill->next->prev = kill->prev;
			e->curwin->file = kill->next;
		} else {
			kill->prev->next = NULL;
			e->curwin->file = kill->prev;
		}
	} else {
		if( kill->next ) {
			kill->next->prev = NULL;
			e->curwin->file = kill->next;
		} else e->curwin->file = NULL;
	}

	for( w = e->curwin; w->prev; w = w->prev );
	for( ; w; ) {
		w2 = w->next;
		if( w->file == kill ) {
			if( w->next || w->prev ) {
				oldwin = e->curwin;
				e->curwin = w;
				do_closewin( "" );
				e->curwin = oldwin;
			} else {
				w->file = e->curwin->file;
				if( w->file ) w->curline = w->file->firstline;
				else w->curline = NULL;
				w->line = 0;
				w->col = 0;
			}
		}
		w = w2;
	}
	
	fclose( kill->orig );
	free( kill );

	return 1;
}


/*
 * tries to kill the current line
 */
int do_delline( char *arg )
{
	STR_LINE	*kill;
	
	if( !( kill = e->curwin->curline ) ) return 1;
	if( kill->prev ) {
		if( kill->next ) {
			kill->prev->next = kill->next;
			kill->next->prev = kill->prev;
			e->curwin->curline = kill->next;
		} else {
			free( kill->buf );
			kill->size = 0;
			return 1;
		}
	} else {
		if( kill->next ) {
			kill->next->prev = NULL;
			e->curwin->curline = kill->next;
		} else e->curwin->curline = NULL;
	}

	free( kill->buf );
	free( kill );

	return 1;
}


/*
 * deletes a character to the left of the cursor
 */
int do_delleft( char *arg )
{
	int		i, col;
	STR_LINE	*l;
	
	if( ( col = do_cursoroffset( "" ) - 1 ) == -1 ) return 0;
	if( !( l = e->curwin->curline ) ) return 0;
	
	if( !readlinetomem() ) return 0;
	
	if( !col ) return 1;
	
	for( i = col - 1; i < l->size; i++ )
		l->buf[i] = l->buf[i + 1];
		
	e->curwin->col--;
	if( e->curwin->leftcol > e->curwin->col )
		e->curwin->leftcol--;
	l->size--;
	
	do_refreshline( "" );
	do_refreshcursor( "" );
	
	return 1;
}


/*
 * deletes a character to the right of the cursor
 */
int do_delright( char *arg )
{
	int		i, col;
	STR_LINE	*l;
	
	if( ( col = do_cursoroffset( "" ) - 1 ) == -1 ) return 0;
	if( !( l = e->curwin->curline ) ) return 0;
	
	if( !readlinetomem() ) return 0;
	
	if( col >= l->size - 1 ) return 1;
	
	for( i = col; i < l->size; i++ )
		l->buf[i] = l->buf[i + 1];
		
	l->size--;
	
	do_refreshline( "" );
	do_refreshcursor( "" );
	
	return 1;
}


/*
 * inserts a string (of one or more chars) at the current position
 */
int do_insert( char *arg )
{
	int		i, col, length, noerror;
	STR_LINE	*l, *newline;
	char		*buf;
	
	if( ( col = do_cursoroffset( "" ) - 1 ) == -1 ) return 0;
	if( !( l = e->curwin->curline ) ) return 0;
	
	if( !( length = strlen( arg ) ) ) return 1;
	
	if( !readlinetomem() ) return 0;
	
	if( l->maxsize < l->size + length ) {
		i = ( (int) ( ( l->size + length ) / ALLOC_SIZE ) + 1 ) * ALLOC_SIZE;
		if( !( buf = (char *) realloc( l->buf, i ) ) ) return 0;
		l->maxsize = i;
		l->buf = buf;
	}
	
	for( i = l->size; i >= col; i-- )
		l->buf[i + length] = l->buf[i];
	for( i = col; i < col + length; i++ ) {
		l->buf[i] = arg[i - col];
		if( l->buf[i] == '\t' ) 
			e->curwin->col = ( ( e->curwin->col / e->curwin->file->tabsize ) + 1 ) * e->curwin->file->tabsize;
		else if( l->buf[i] == '\n' && i < l->size ) {
			if( !( newline = (STR_LINE *) e_alloc( sizeof( STR_LINE ) ) ) ) {
				for( i = l->size; i >= col; i-- )
					l->buf[i] = l->buf[i + length];
				do_showstatus( "error: out of memory!" );
				return 0;
			}
			newline->offset = 0;
			newline->size = 0;
			newline->maxsize = 0;
			newline->buf = NULL;
			newline->bufpos = 0;
			
			e->curwin->curline = newline;
			noerror = do_insert( l->buf + i + 1 ); 
			if( !noerror ) {
				e->curwin->curline = l;
				do_showstatus( "error: couldn't add new line!" );
				free( newline );
				return 0;
			}
			if( l->next ) {
				l->next->prev = newline;
				newline->next = l->next;
			} else
				newline->next = NULL;
			l->next = newline;
			newline->prev = l;
			e->curwin->line++;
			do_startofline( "" );
			l->buf[i + 1] = '\0';
			l->size = strlen( l->buf );
			do_refresh( "" );
		} else
			e->curwin->col++;
	}
	l->size = l->size + length;
	
	if( e->curwin->leftcol + e->curwin->width <= e->curwin->col ) {
		e->curwin->leftcol = e->curwin->width - e->curwin->col + 1;
		do_refresh( "" );
	} else {
		do_refreshline( "" );
		do_refreshcursor( "" );
	}
	
	return 1;
}


/*
 * tries to close the current window
 */
int do_closewin( char *arg )
{
	STR_WIN 	*kill;
	
	if( !( kill = e->curwin ) ) return 1;
	
	if( kill->prev ) {
		if( kill->next ) {
			kill->prev->next = kill->next;
			kill->next->prev = kill->prev;
			e->curwin = kill->next;
		} else {
			kill->prev->next = NULL;
			e->curwin = kill->prev;
		}
	} else {
		if( kill->next ) {
			kill->next->prev = NULL;
			e->curwin = kill->next;
		} else e->curwin = NULL;
	}
	
	free( kill );
	
	if( !e->curwin ) do_exit( "0" );
	
	return 1;
}


/*
 * tries to open a new window
 */
int do_newwin( char *arg )
{
	STR_WIN		*newwin;

	if( !( newwin = (STR_WIN *) e_alloc( sizeof( STR_WIN ) ) ) ) return 0;
	newwin->file = NULL;
	newwin->curline = NULL;
	newwin->line = 0;
	newwin->col = 0;
	newwin->topline = 0;
	newwin->leftcol = 0;
	
	if( e->curwin ) {
		if( e->curwin->next ) {
			newwin->next = e->curwin->next;
			newwin->next->prev = newwin;
		} else newwin->next = NULL;
		e->curwin->next = newwin;
		newwin->prev = e->curwin;
		newwin->xpos = e->curwin->xpos;
		newwin->width = e->curwin->width;
		newwin->ypos = e->curwin->ypos + ( e->curwin->height >> 1 );
		e->curwin->height >>= 1;
		newwin->height = e->curwin->height;
	} else {
		e->curwin = newwin;
		newwin->xpos = 0;
		newwin->ypos = 0;
		newwin->width = COLS;
		newwin->height = LINES;
		newwin->prev = NULL;
		newwin->next = NULL;
	}
	
	return 1;
}


/*
 * refreshes the screen with any waiting updates
 */
int do_refresh( char *arg )
{
	int		i, oldline;
	STR_LINE	*oldcurline;
	
	oldcurline = e->curwin->curline;
	oldline = e->curwin->line;
	
	for( i = e->curwin->line; i > e->curwin->topline; i-- )
		e->curwin->curline = e->curwin->curline->prev;
		
	e->curwin->line = e->curwin->topline;
	
	for( i = 0; i < e->curwin->height; i++ ) {
		if( !e->curwin->curline ) {
			clrtobot();
			break;
		}
		else {
			do_refreshline( "" );
			e->curwin->curline = e->curwin->curline->next;
			e->curwin->line++;
		}
	}
	
	e->curwin->curline = oldcurline;
	e->curwin->line = oldline;
	
	do_refreshcursor( "" );
	
	return 1;
}


/*
 * refreshes the current line with any waiting updates
 */
int do_refreshline( char *arg )
{
	int		col, oldcol, i, i2;
	char		c;
	STR_LINE	*l;
	
	if( !( l = e->curwin->curline ) ) return 0;
	
	move( e->curwin->ypos + e->curwin->line - e->curwin->topline, 0 );

	if( e->curwin->curline->bufpos == 1 ) fseek( e->curwin->file->orig, l->offset, SEEK_SET );
	
	for( i = 0, col = 0; i < l->size; i++ ) {
		if( l->bufpos == 0 ) c = l->buf[i];
		else if( l->bufpos == 1 ) c = fgetc( e->curwin->file->orig );
		else if( l->bufpos == 2 ) {
			e_msg( "swapping's not enabled yet..." );
			return 0;
		}
		if( c == '\t' ) {
			oldcol = col;
			col = ( ( col / e->curwin->file->tabsize ) + 1 ) * e->curwin->file->tabsize;
			if( col > e->curwin->leftcol ) {
				if( oldcol >= e->curwin->leftcol ) {
					for( i2 = oldcol; i2 < col; i2++ )
						addch( ' ' );
				} else {
					for( i2 = e->curwin->leftcol; i2 < col; i2++ )
						addch( ' ' );
				}
			}
		} else if( c == '\n' ) {
			clrtoeol();
			break;
		} else {
			if( col >= e->curwin->leftcol )
				addch( c );
			col++;
		}
		if( col >= e->curwin->leftcol + e->curwin->width )
			break;
	}
	
	return 1;
}


/*
 * outputs a message to the bottomline (or just to the screen if we're not
 * inside the editor. A '\n' is appended automagically if we're not inside 
 * the editor. If the string is of zero length then the statusline is removed
 */
int do_showstatus( char *arg )
{
	STR_LINE	*oldl;
	int		oldline;
	static int	msglines = 0;
	
	int		x, y, i;

	if( !arg ) return 1;

	if( !strlen( arg ) ) {
		if( !msglines )
			return 1;
		oldline = e->curwin->line;
		oldl = e->curwin->curline;

		for( ; e->curwin->line >= e->curwin->topline + e->curwin->height - msglines; e->curwin->line-- )
			if( !( e->curwin->curline = e->curwin->curline->prev ) ) break;

		for( ; e->curwin->line < e->curwin->topline + e->curwin->height - msglines && e->curwin->curline; e->curwin->line++ )
			e->curwin->curline = e->curwin->curline->next;

		if( e->curwin->line < e->curwin->topline + e->curwin->height - msglines ) {
			move( LINES - msglines, 0 );
			clrtobot();
		} else
			for( i = 0; ; ) {
				do_refreshline( "" );
				if( ++i >= msglines )
					break;
				e->curwin->line++;
				if( !( e->curwin->curline = e->curwin->curline->next ) ) {
					do_refreshcursor( "" );
					clrtobot();
					break;
				}
			}
		e->curwin->line = oldline;
		e->curwin->curline = oldl;
		do_refreshcursor( "" );
		msglines = 0;
		return 1;
	}

	arg = stripspaces( arg );

	if( e->status ) {
		getyx( stdscr, y, x );
		msglines = strlen( arg ) / COLS + 1;
		mvprintw( LINES - msglines, 0, "%s", arg );
		clrtoeol();
		move( y, x );
	} else printf( "%s\n", arg );
	refresh();
	
	return 1;
}


/*
 * puts the physical cursor where it should be
 */
int do_refreshcursor( char *arg )
{
	move( e->curwin->ypos + e->curwin->line - e->curwin->topline, e->curwin->xpos + e->curwin->col - e->curwin->leftcol );
	
	return 1;
}


/*
 * moves the cursor one line up
 *
 * returns: 0 on error, 1 on success, 2 if already at first line
 */
int do_up( char *arg )
{
	if( !e->curwin->curline->prev ) return 2;
	
	e->curwin->curline = e->curwin->curline->prev;
	e->curwin->line--;
	if( e->curwin->topline > e->curwin->line ) {
		e->curwin->topline--;
		scrl( -1 );
		do_refreshline( "" );
	}
	do_refreshcursor( "" );
	
	return 1;
}


/*
 * moves the cursor one line down
 *
 * returns: 0 on error, 1 on success, 2 if already at last line
 */
int do_down( char *arg )
{
	if( !e->curwin->curline->next ) return 2;
	
	e->curwin->curline = e->curwin->curline->next;
	e->curwin->line++;
	if( e->curwin->topline + e->curwin->height <= e->curwin->line ) {
		e->curwin->topline++;
		scrl( 1 );
		do_refreshline( "" );
	}
	do_refreshcursor( "" );
		
	return 1;
}


/*
 * moves the cursor one column to the left
 *
 * returns: 0 on error, 1 on success, 2 if already at first column
 */
int do_left( char *arg )
{
	int		col, oldcol, i;
	char		c;
	STR_LINE	*l;
	
	if( !e->curwin->col ) {
		if( e->curwin->line ) {
			do_up( "" );
			do_endofline( "" );
		}
		return 2;
	}
	
	if( !( l = e->curwin->curline ) ) return 0;
	
	if( e->curwin->curline->bufpos == 1 ) fseek( e->curwin->file->orig, l->offset, SEEK_SET );
	
	for( i = 0, col = 0; i < l->size; i++ ) {
		if( l->bufpos == 0 ) c = l->buf[i];
		else if( l->bufpos == 1 ) c = fgetc( e->curwin->file->orig );
		else if( l->bufpos == 2 ) {
			e_msg( "swapping's not enabled yet..." );
			return 0;
		}
		if( c == '\n' ) break;
		oldcol = col;
		if( c == '\t' ) col = ( ( col / e->curwin->file->tabsize ) + 1 ) * e->curwin->file->tabsize;
		else col++;
		if( col >= e->curwin->col ) break;
	}

	if( c == '\n' ) {
		do_endofline( "" );
		return 2;
	}
	
	e->curwin->col = oldcol;
	if( e->curwin->leftcol > e->curwin->col ) {
		e->curwin->leftcol = e->curwin->col;
		do_refresh( "" );
	} else
		do_refreshcursor( "" );

	return 1;
}


/*
 * moves the cursor one character to the right
 *
 * returns: 0 on error, 1 on success, 2 if already at last column, 3 if at eof
 */
int do_right( char *arg )
{
	int		col, i;
	char		c;
	STR_LINE	*l;
	
	if( !( l = e->curwin->curline ) ) return 0;
	
	if( e->curwin->curline->bufpos == 1 ) fseek( e->curwin->file->orig, l->offset, SEEK_SET );
	
	for( i = 0, col = 0; i < l->size; i++ ) {
		if( l->bufpos == 0 ) c = l->buf[i];
		else if( l->bufpos == 1 ) c = fgetc( e->curwin->file->orig );
		else if( l->bufpos == 2 ) {
			e_msg( "swapping's not enabled yet..." );
			return 0;
		}
		if( c == '\n' ) break;
		if( c == '\t' ) col = ( ( col / e->curwin->file->tabsize ) + 1 ) * e->curwin->file->tabsize;
		else col++;
		if( col > e->curwin->col ) break;
	}

	if( c == '\n' ) {
		if( l->next ) {
			do_startofline( "" );
			do_down( "" );
		}
		return 2;
	}
	
	e->curwin->col = col;
	if( e->curwin->leftcol + e->curwin->width <= e->curwin->col ) {
		e->curwin->leftcol = e->curwin->col - e->curwin->width + 1;
		do_refresh( "" );
	} else
		do_refreshcursor( "" );
		
	return 1;
}


/*
 * moves the cursor one page up
 *
 * returns: 0 on error, 1 on success, 2 if already at last line
 */
int do_pageup( char *arg )
{
	int		i;
	STR_LINE	*l;
	
	if( !( l = e->curwin->curline ) ) return 1;
	
	for( i = 0; i < e->curwin->height / 2; i++ ) {
		if( !( l = l->prev ) ) break;
		if( e->curwin->topline ) e->curwin->topline--;
		e->curwin->line--;
		e->curwin->curline = l;
	}

	do_refresh( "" );
	
	return 1;
}


/*
 * moves the cursor one page down
 */
int do_pagedown( char *arg )
{
	int		i;
	STR_LINE	*l;
	
	if( !( l = e->curwin->curline ) ) return 1;
	
	for( i = 0; i < e->curwin->height / 2; i++ ) {
		if( !( l = l->next ) ) break;
		e->curwin->topline++;
		e->curwin->line++;
		e->curwin->curline = l;
	}
	
	do_refresh( "" );
	
	return 1;
}


/*
 * returns the ( byteoffset + 1 ) of the cursor in the current STR_LINE
 * please note the +1, this is done to maintain the "0 means error"-standard
 */
int do_cursoroffset( char *arg )
{
	int		col, i;
	char		c;
	STR_LINE	*l;
	
	l = e->curwin->curline;
	
	if( e->curwin->curline->bufpos == 1 ) fseek( e->curwin->file->orig, l->offset, SEEK_SET );
	
	for( i = 0, col = 0; i < l->size; i++ ) {
		if( l->bufpos == 0 ) c = l->buf[i];
		else if( l->bufpos == 1 ) c = fgetc( e->curwin->file->orig );
		else if( l->bufpos == 2 ) {
			e_msg( "swapping's not enabled yet..." );
			return 0;
		}
		if( c == '\n' ) break;
		if( c == '\t' ) col = ( ( col / e->curwin->file->tabsize ) + 1 ) * e->curwin->file->tabsize;
		else col++;
		if( col > e->curwin->col ) break;
	}
	
	return i + 1;
}


/*
 * moves the cursor to the start of the line
 */
int do_startofline( char *arg )
{
	int		oldleftcol;
	
	oldleftcol = e->curwin->leftcol;
	
	e->curwin->col = 0;
	e->curwin->leftcol = 0;
	
	if( oldleftcol )
		do_refresh( "" );
	else
		do_refreshcursor( "" );

	return 1;
}


/*
 * moves the cursor to the end of the line
 */
int do_endofline( char *arg )
{
	int		i;
	char		c;

	if( e->curwin->curline->bufpos == 2 ) {
		e_msg( "swapping's not enabled yet, you're not supposed to be getting this msg!" );
		return 0;
	}

	if( e->curwin->curline->bufpos == 1 ) fseek( e->curwin->file->orig, e->curwin->curline->offset, SEEK_SET );

	e->curwin->col = 0;
	for( i = 0; i < e->curwin->curline->size; i++ ) {
		if( e->curwin->curline->bufpos == 0 ) c = e->curwin->curline->buf[i];
		else if( e->curwin->curline->bufpos == 1 ) c = fgetc( e->curwin->file->orig );
		if( c == '\n' ) break;
		if( c == '\t' ) e->curwin->col = ( ( e->curwin->col / e->curwin->file->tabsize ) + 1 ) * e->curwin->file->tabsize;
		else e->curwin->col++;
	}
	if( e->curwin->leftcol + e->curwin->width <= e->curwin->col ) {
		e->curwin->leftcol = e->curwin->col - e->curwin->width + 1;
		do_refresh( "" );
	} else
		do_refreshcursor( "" );
	
	return 1;
}


/*
 * makes the next file the current file
 */
int do_nextfile( char *arg )
{
	e->curwin->file->curline = e->curwin->curline;
	e->curwin->file->line = e->curwin->line;
	e->curwin->file->col = e->curwin->col;
	e->curwin->file->topline = e->curwin->topline;	
	e->curwin->file->leftcol = e->curwin->leftcol;	
	if( e->curwin->file->next )
		e->curwin->file = e->curwin->file->next;
	else
		for( ; e->curwin->file->prev; e->curwin->file = e->curwin->file->prev )
			;

	e->curwin->curline = e->curwin->file->curline;
	e->curwin->line = e->curwin->file->line;
	e->curwin->topline = e->curwin->file->topline;
	e->curwin->col = e->curwin->file->col;
	e->curwin->leftcol = e->curwin->file->leftcol;
	
	do_refresh( "" );
	
	return 1;
}


/*
 * makes the previous file the current file
 */
int do_prevfile( char *arg )
{
	e->curwin->file->curline = e->curwin->curline;
	e->curwin->file->line = e->curwin->line;
	e->curwin->file->col = e->curwin->col;
	e->curwin->file->topline = e->curwin->topline;	
	e->curwin->file->leftcol = e->curwin->leftcol;	
	if( e->curwin->file->prev )
		e->curwin->file = e->curwin->file->prev;
	else
		for( ; e->curwin->file->next; e->curwin->file = e->curwin->file->next )
			;

	e->curwin->curline = e->curwin->file->curline;
	e->curwin->line = e->curwin->file->line;
	e->curwin->topline = e->curwin->file->topline;
	e->curwin->col = e->curwin->file->col;
	e->curwin->leftcol = e->curwin->file->leftcol;
	
	do_refresh( "" );
	
	return 1;
}


/*
 * saves the current file
 */
int do_savefile( char *arg )
{
	STR_LINE		*l;
	char			tmpline[MAX_LINESIZE];
	FILE			*outfile;
	char			*tmpfilename;

	if( !( tmpfilename = tmpnam( NULL ) ) ) {
		e_msg( "error: couldn't generate name for temporary file!" );
		return 0;
	}

	if( !( outfile = fopen( tmpfilename, "wb" ) ) ) {
		e_msg( "error: couldn't open temporary file '%s' for writing!", tmpfilename );
		return 0;
	}
	
	for( l = e->curwin->file->firstline; l; l = l->next ) {
		if( !l->size )
			continue;
		if( !l->bufpos )
			fputs( l->buf, outfile );
		else if( l->bufpos == 1 ) {
			fseek( e->curwin->file->orig, l->offset, SEEK_SET );
			if( !( fgets( tmpline, MAX_LINESIZE, e->curwin->file->orig ) ) ) {
				e_msg( "error: couldn't read line from original file!" );
				fclose( outfile );
				remove( tmpfilename );
				return 0;
			}
			fputs( tmpline, outfile );
		} else if( l->bufpos == 2 ) {
			e_msg( "error: swapping's not supported yet!" );
			fclose( outfile );
			remove( tmpfilename );
			return 0;
		}			
	}
	
	fclose( outfile );

	if( rename( tmpfilename, e->curwin->file->name ) ) {
		e_msg( "error: couldn't rename your newly saved file '%s' to '%s'! Sorry for any inconvenience...", tmpfilename, e->curwin->file->name );
		getch();
		return 0;
	}

	remove( tmpfilename );
	
	e_msg( "file '%s' was successfully saved!", e->curwin->file->name );
	
	return 1;
}


/*
 * refreshes the statusbar with any waiting updates
 */
int do_refreshstatusbar( char *arg )
{
	int		line;
	char		tmp[MAX_STATUSBARSIZE];
	
	line = atoi( e->statusbarpos + 1 );
	
	if( e->statusbarpos[0] == '-' )
		line = LINES - 1 - line;
	
	if( !parse_statusbar_definition( e->statusbarleft, tmp ) )
		return 0;
	
	mvprintw( line, 0, "%s", tmp );
	
	do_refreshcursor( "" );
	
	return 1;
}


/*
 * parses a statusbar-definition ( s ) into a regular string ( s2 )
 */
int parse_statusbar_definition( char *s, char *s2 )
{
	int		i, i2, code;
	char		c;
	
	s2[0] = '\0';
	
	for( i = 0, i2 = 0, code = 0; i < strlen( s ) && i2 < MAX_STATUSBARSIZE; i++ ) {
		c = s[i];
		if( c != '\\' && !code ) {
			s2[i2] = c;
			code = 0;
		} else if( c == '\\' ) 
			code = 1;
		else {
			switch( c ) {
				case 'p':
					strncat( s2, e->curwin->file->name, MAX_STATUSBARSIZE - strlen( s2 ) );
					break;
				default:
					break;
			}
			code = 0;
		}
	}
	
	if( i2 >= MAX_STATUSBARSIZE ) {
		e_msg( "error: couldn't parse statusbar-definition! String too long. You should increase MAX_STATUSBARSIZE in e.h and recompile." );
		return 0;
	}
	
	return 1;
}
